# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).

"""Service Double implementation.

The ServiceMock class in this file is used at test-run-time to mock out a call
to a remote service API view.
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import *  # NOQA

import functools
import json

from future.moves.urllib.parse import urljoin
from fixtures import Fixture
import responses

from acceptable._validation import validate
from acceptable.mocks import responses_manager

def service_mock(service, methods, url, input_schema, output_schema):
    return functools.partial(
        ServiceMock,
        service, methods, url, input_schema, output_schema)


SERVICE_LOCATIONS = {}


def set_service_locations(service_locations):
    global SERVICE_LOCATIONS
    SERVICE_LOCATIONS = service_locations


def get_service_locations():
    global SERVICE_LOCATIONS
    return SERVICE_LOCATIONS


class ServiceMock(Fixture):
    # Kept for backwards compatibility
    _requests_mock = responses.mock

    def __init__(self, service, methods, url, input_schema, output_schema,
                 output, output_status=200):
        super().__init__()
        self._service = service
        self._methods = methods
        self._url = url
        self._input_schema = input_schema
        self._output_schema = output_schema
        self._output = output
        self._output_status = output_status

    def _setUp(self):
        if self._output_schema:
            error_list = validate(self._output, self._output_schema)
            if error_list:
                msg = (
                    "While setting up a service mock for the '{s._service}' "
                    "service's '{s._url}' endpoint, the specified output "
                    "does not match the service's endpoint output schema.\n\n"
                    "The errors are:\n{errors}\n\n"
                ).format(s=self, errors='\n'.join(error_list))
                raise AssertionError(msg)

        config = get_service_locations()
        service_location = config.get(self._service)
        if service_location is None:
            raise AssertionError(
                "A service mock for the '%s' service was requested, but the "
                "mock has not been configured with a location for that "
                "service. Ensure set_service_locations has been "
                "called before the mock is required, and that the locations "
                "dictionary contains a key for the '%s' service."
                % (self._service, self._service)
            )

        full_url = urljoin(service_location, self._url)

        def _callback(request):
            if self._input_schema:
                payload = json.loads(request.body.decode())
                error_list = validate(payload, self._input_schema)
                if error_list:
                    # TODO: raise AssertionError here, since this is in a test?
                    return (
                        400,
                        {'Content-Type': 'application/json'},
                        json.dumps(error_list),
                    )
            # TODO: Do we need to support more than just json responses?
            return (
                self._output_status,
                {"Content-Type": "application/json"},
                json.dumps(self._output)
            )

        responses_manager.attach()
        self.addCleanup(responses_manager.detach)
        for method in self._methods:
            responses.mock.add_callback(method, full_url, _callback)

    @property
    def calls(self):
        return responses.mock.calls



# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).


# including this causes issues with type types in __all__, which need to be
# native strings, so we explicitly do not want unicode literals here
# from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from ._service import (  # NOQA
    AcceptableService,
    get_metadata,
)
from ._validation import (  # NOQA
    DataValidationError,
    validate_body,
    validate_output,
)


# __all__ strings must be bytes in py2 and unicode in py3
__all__ = [
    'AcceptableService',
    'DataValidationError',
    'get_metadata',
    'validate_body',
    'validate_output',
]



# Copyright 2019 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).
from __future__ import print_function
from collections import defaultdict
import json
import textwrap
from sys import stdout


def write_service_begin(stream, service_name):
    begin = (
        u"\n"
        u"{service_name} = ServiceFactory(\n"
        u"        {service_name!r},\n"
        u"        [\n"
    ).format(
        service_name=service_name
    )
    stream.write(begin)


def write_service_end(stream):
    stream.write(
        u"    ]\n"
        u")\n"
        u"\n"
    )


def write_endpoint_spec(stream, api_data):
    endpoint = (
        u"            EndpointSpec(\n"
        u"                {api_name!r},\n"
        u"                {url!r},\n"
        u"                {methods!r},\n"
        u"                {request_schema!r},\n"
        u"                {response_schema!r}\n"
        u"            ),\n"
    ).format(**api_data)
    stream.write(endpoint)


HEADER_TEXT = u"""\
# This file is AUTO GENERATED. Do not edit this file directly. Instead,
# re-generate it.

from acceptable.mocks import ServiceFactory, EndpointSpec

"""


def write_header(stream):
    stream.write(HEADER_TEXT)


def generate_service_factory(metadata, stream=stdout):
    write_header(stream)
    services = defaultdict(list)
    for module_data in metadata.values():
        if not isinstance(module_data, dict):
            continue
        for api_data in module_data['apis'].values():
            services[api_data['service']].append(api_data)
    for service_name, apis in services.items():
        write_service_begin(stream, service_name)
        for api in apis:
            write_endpoint_spec(stream, api)
        write_service_end(stream)





# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import *  # NOQA
__metaclass__ = type

from enum import IntEnum
import os


class LintTypes(IntEnum):
    WARNING = 0
    DOCUMENTATION = 1
    ERROR = 2


# shortcuts
WARNING = LintTypes.WARNING
DOCUMENTATION = LintTypes.DOCUMENTATION
ERROR = LintTypes.ERROR


class Message():
    """A linter message to the user."""
    level = None

    def __init__(self, name, msg, *args, **kwargs):
        self.name = name
        self.location = kwargs.pop('location', None)
        self.api_name = kwargs.pop('api_name', None)
        self.msg = msg.format(*args, **kwargs)

    def __str__(self):
        output = '{}: API {} at {}: {}'.format(
            self.level.name.title(),
            self.api_name,
            self.name,
            self.msg,
        )

        if self.location is None:
            return output
        else:
            return '{}:{}: {}'.format(
                os.path.relpath(self.location['filename']),
                self.location['lineno'],
                output,
            )


class LintError(Message):
    level = ERROR


class LintWarning(Message):
    level = WARNING


class LintFixit(Message):
    level = DOCUMENTATION


class CheckChangelog(Message):
    def __init__(self, name, revision):
        super().__init__(name, '')
        self.revision = revision


def metadata_lint(old, new, locations):
    """Run the linter over the new metadata, comparing to the old."""
    # ensure we don't modify the metadata
    old = old.copy()
    new = new.copy()
    # remove version info
    old.pop('$version', None)
    new.pop('$version', None)

    for old_group_name in old:
        if old_group_name not in new:
            yield LintError('', 'api group removed', api_name=old_group_name)

    for group_name, new_group in new.items():
        old_group = old.get(group_name, {'apis': {}})

        for name, api in new_group['apis'].items():
            old_api = old_group['apis'].get(name, {})
            api_locations = locations[name]
            for message in lint_api(name, old_api, api, api_locations):
                message.api_name = name
                if message.location is None:
                    message.location = api_locations['api']
                yield message


def lint_api(api_name, old, new, locations):
    """Lint an acceptable api metadata."""
    is_new_api = not old
    api_location = locations['api']
    changelog = new.get('changelog', {})
    changelog_location = api_location

    if locations['changelog']:
        changelog_location = list(locations['changelog'].values())[0]

    # apis must have documentation if they are new
    if not new.get('doc'):
        msg_type = LintError if is_new_api else LintWarning
        yield msg_type(
            'doc',
            'missing docstring documentation',
            api_name=api_name,
            location=locations.get('view', api_location)
        )

    introduced_at = new.get('introduced_at')
    if introduced_at is None:
        yield LintError(
            'introduced_at',
            'missing introduced_at field',
            location=api_location,
        )

    if not is_new_api:
        # cannot change introduced_at if we already have it
        old_introduced_at = old.get('introduced_at')
        if old_introduced_at is not None:
            if old_introduced_at != introduced_at:
                yield LintError(
                    'introduced_at',
                    'introduced_at changed from {} to {}',
                    old_introduced_at,
                    introduced_at,
                    api_name=api_name,
                    location=api_location,
                )

    # cannot change url
    if new['url'] != old.get('url', new['url']):
        yield LintError(
            'url',
            'url changed from {} to {}',
            old['url'],
            new['url'],
            api_name=api_name,
            location=api_location,
        )

    # cannot add required fields
    for removed in set(old.get('methods', [])) - set(new['methods']):
        yield LintError(
            'methods',
            'HTTP method {} removed',
            removed,
            api_name=api_name,
            location=api_location,
        )

    for schema in ['request_schema', 'response_schema']:
        new_schema = new.get(schema)
        if new_schema is None:
            if old.get(schema) is not None:
                yield LintError(
                    schema,
                    '{} schema removed',
                    schema.split('_')[0].title(),
                    api_name=api_name,
                    location=api_location,
                )
            continue

        schema_location = locations[schema]
        old_schema = old.get(schema) or {}

        for message in walk_schema(
                schema, old_schema, new_schema, root=True, new_api=is_new_api):
            if isinstance(message, CheckChangelog):
                if message.revision not in changelog:
                    yield LintFixit(
                        message.name,
                        'No changelog entry for revision {}',
                        message.revision,
                        location=changelog_location,
                    )
            else:
                # add in here, saves passing it down the recursive call
                message.location = schema_location
                yield message


def get_schema_types(schema):
    schema_type = schema.get('type')
    if schema_type is None:
        return []
    elif isinstance(schema_type, (str, bytes)):
        return [schema_type]
    else:
        return schema_type


def check_custom_attrs(name, old, new, new_api=False):
    # these are our custom schema properties, not in the jsonschema
    # standard, and we don't need them on the root schema object, as that
    # takes its doc from the function docstring and its introduced_at from
    # the api definition
    if 'description' not in new:
        yield LintWarning(name + '.description', 'missing description field')

    if not new_api:
        introduced_at = new.get('introduced_at')
        if introduced_at is None:
            if not old:  # new field
                yield LintFixit(
                    name + '.introduced_at', 'missing introduced_at field')
        else:
            introduced_at_changed = False
            old_introduced_at = old.get('introduced_at')
            if old_introduced_at is not None:
                introduced_at_changed = old_introduced_at != introduced_at

            if introduced_at_changed:
                yield LintError(
                    name + '.introduced_at',
                    'introduced_at changed from {} to {}',
                    old_introduced_at,
                    introduced_at,
                )
            else:
                # we have a specific introduced_at field, make sure to
                # check it's referenced by a changelog entry
                yield CheckChangelog(name, introduced_at)


def walk_schema(name, old, new, root=False, new_api=False):
    if not root:
        for i in check_custom_attrs(name, old, new, new_api):
            yield i

    types = get_schema_types(new)
    old_types = get_schema_types(old)
    for removed in set(old_types) - set(types):
        yield LintError(
            name + '.type',
            'cannot remove type {} from field',
            removed,
        )

    # you cannot add new required fields to an existing API.
    if not new_api:
        old_required = old.get('required', [])
        for removed in set(new.get('required', [])) - set(old_required):
            yield LintError(
                name + '.required',
                'Cannot require new field {}',
                removed,
            )

    if 'object' in types:
        properties = new.get('properties', {})
        old_properties = old.get('properties', {})

        for deleted in set(old_properties) - set(properties):
            yield LintError(
                name + '.' + deleted, 'cannot delete field {}', deleted)

        for prop, value in sorted(properties.items()):
            for i in walk_schema(
                    name + '.' + prop,
                    old_properties.get(prop, {}),
                    value,
                    new_api=new_api):
                yield i

    if 'array' in types and 'items' in new:
        for i in walk_schema(
                name + '.items',
                old.get('items', {}),
                new['items'],
                new_api=new_api):
            yield i



# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import *  # NOQA

import json
import functools

import jsonschema

from acceptable.util import get_callsite_location, sort_schema


class DataValidationError(Exception):
    """Raises when a request body fails validation."""

    def __init__(self, error_list):
        self.error_list = error_list

    def __repr__(self):
        return "DataValidationError: %s" % ", ".join(self.error_list)

    def __str__(self):
        return repr(self)


def validate_body(schema):
    """Validate the body of incoming requests for a flask view.

    An example usage might look like this::

        from snapstore_schemas import validate_body


        @validate_body({
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'snap_id': {'type': 'string'},
                    'series': {'type': 'string'},
                    'name': {'type': 'string'},
                    'title': {'type': 'string'},
                    'keywords': {
                        'type': 'array',
                        'items': {'type': 'string'}
                    },
                    'summary': {'type': 'string'},
                    'description': {'type': 'string'},
                    'created_at': {'type': 'string'},
                },
                'required': ['snap_id', 'series'],
                'additionalProperties': False
            }
        })
        def my_flask_view():
            # view code here
            return "Hello World", 200

    All incoming request that have been routed to this view will be matched
    against the specified schema. If the request body does not match the schema
    an instance of `DataValidationError` will be raised.

    By default this will cause the flask application to return a 500 response,
    but this can be customised by telling flask how to handle these exceptions.
    The exception instance has an 'error_list' attribute that contains a list
    of all the errors encountered while processing the request body.
    """
    location = get_callsite_location()

    def decorator(fn):
        validate_schema(schema)
        wrapper = wrap_request(fn, schema)
        record_schemas(
            fn, wrapper, location, request_schema=sort_schema(schema))
        return wrapper

    return decorator


def wrap_request(fn, schema):
    from flask import request

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        payload = request.get_json(silent=True, cache=True, force=True)
        # If flask can't parse the payload, we want to return a sensible
        # error message, so we try and parse it ourselves. Setting silent
        # to False above isn't good enough, as the generated error message
        # is not informative enough.
        if payload is None:
            try:
                payload = json.loads(request.data.decode(request.charset))
            except ValueError as e:
                raise DataValidationError([
                    "Error decoding JSON request body: %s" % str(e)])
        error_list = validate(payload, schema)
        if error_list:
            raise DataValidationError(error_list)
        return fn(*args, **kwargs)

    return wrapper


def record_schemas(
        fn, wrapper, location, request_schema=None, response_schema=None):
    """Support extracting the schema from the decorated function."""
    # have we already been decorated by an acceptable api call?
    has_acceptable = hasattr(fn, '_acceptable_metadata')

    if request_schema is not None:
        # preserve schema for later use
        wrapper._request_schema = wrapper._request_schema = request_schema
        wrapper._request_schema_location = location
        if has_acceptable:
            fn._acceptable_metadata._request_schema = request_schema
            fn._acceptable_metadata._request_schema_location = location

    if response_schema is not None:
        # preserve schema for later use
        wrapper._response_schema = wrapper._response_schema = response_schema
        wrapper._response_schema_location = location
        if has_acceptable:
            fn._acceptable_metadata._response_schema = response_schema
            fn._acceptable_metadata._response_schema_location = location


def validate_output(schema):
    """Validate the body of a response from a flask view.

    Like `validate_body`, this function compares a json document to a
    jsonschema specification. However, this function applies the schema to the
    view response.

    Instead of the view returning a flask response object, it should instead
    return a Python list or dictionary. For example::

        from snapstore_schemas import validate_output

        @validate_output({
            'type': 'object',
            'properties': {
                'ok': {'type': 'boolean'},
            },
            'required': ['ok'],
            'additionalProperties': False
        }
        def my_flask_view():
            # view code here
            return {'ok': True}

    Every view response will be evaluated against the schema. Any that do not
    comply with the schema will cause DataValidationError to be raised.
    """
    location = get_callsite_location()

    def decorator(fn):
        validate_schema(schema)
        wrapper = wrap_response(fn, schema)
        record_schemas(
            fn, wrapper, location, response_schema=sort_schema(schema))
        return wrapper

    return decorator


def wrap_response(fn, schema):
    from flask import current_app, jsonify

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        result = fn(*args, **kwargs)
        if isinstance(result, tuple):
            resp = result[0]
        else:
            resp = result
        if not isinstance(resp, (list, dict)):
            raise ValueError(
                "Unknown response type '%s'. Supported types are list "
                "and dict." % type(resp))

        if current_app.config.get('ACCEPTABLE_VALIDATE_OUTPUT', True):
            error_list = validate(resp, schema)

            assert not error_list,\
                "Response does not comply with output schema: %r.\n%s"\
                % (error_list, resp)

        if isinstance(result, tuple):
            return (jsonify(resp), ) + result[1:]
        else:
            return jsonify(result)
    return wrapper


def validate(payload, schema):
    """Validate `payload` against `schema`, returning an error list.

    jsonschema provides lots of information in it's errors, but it can be a bit
    of work to extract all the information.
    """
    v = jsonschema.Draft4Validator(
        schema, format_checker=jsonschema.FormatChecker())
    error_list = []
    for error in v.iter_errors(payload):
        message = error.message
        location = '/' + '/'.join([str(c) for c in error.absolute_path])
        error_list.append(message + ' at ' + location)
    return error_list


def validate_schema(schema):
    """Validate that 'schema' is correct.

    This validates against the jsonschema v4 draft.

    :raises jsonschema.SchemaError: if the schema is invalid.
    """
    jsonschema.Draft4Validator.check_schema(schema)



# Copyright 2019 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).
from __future__ import absolute_import
from future import standard_library

standard_library.install_aliases()

from collections import namedtuple
import copy
from json import dumps as json_dumps
from json import loads as json_loads
from urllib.parse import urljoin
import responses
from acceptable._validation import validate
import re
from requests.utils import CaseInsensitiveDict

from .responses import responses_manager


class Attrs(object):
    """A utility class allowing the creation of namespaces from a dict.
    Also provides an iterator over the items of the original dict.

    This is used by both Service and ServiceMock to create their 
    endpoints attributes.
    
    e.g:
    a = Attrs(dict(b=1, c=2))
    assert a.b == 1
    assert a.c == 2
    assert dir(a) == ['b', 'c']
    """
    def __init__(self, attrs):
        # I think python name mangling is ok here to help avoid collisions 
        # between instance attributes and names in attrs
        self.__attrs = dict(attrs)
    
    def __dir__(self):
        return list(self.__attrs)
    
    def __getattr__(self, name):
        try:
            return self.__attrs[name]
        except KeyError:
            raise NameError(name)

    def __iter__(self):
        return iter(self.__attrs.items())


Call = namedtuple("Call", "request response error".split())


class CallRecorder(object):
    def __init__(self):
        self._calls = []

    def record(self, mock, request, response, error):
        self._calls.append((mock, Call(request, response, error)))

    def get_calls(self):
        return [c for m, c in self._calls]

    def get_calls_for(self, mock):
        return [c for m, c in self._calls if m == mock]

    def get_calls_for_matching(self, mock, pattern):
        if not hasattr(pattern, "search"):
            pattern = re.compile(pattern)
        return [c for c in self.get_calls_for(mock) if pattern.search(c.request.url)]

    def get_calls_matching(self, pattern):
        if not hasattr(pattern, "search"):
            pattern = re.compile(pattern)
        return [c for m, c in self._calls if pattern.search(c.request.url)]


EndpointSpec = namedtuple(
    "EndpointSpec", ["name", "location", "methods", "request_schema", "response_schema"]
)


VALIDATION_ERROR_TEXT = """
{}:
{!r}

did not match schema:
{!r}

for service {!r} endpoint {!r} on url {!r} errors where:
{!r}
"""


class EndpointMock(object):
    """Provides methods to check calls made to this endpoint mock
    """
    def __init__(
        self,
        call_recorder,
        service_name,
        name,
        methods,
        url,
        request_schema,
        response_schema,
        response_callback,
    ):
        self._call_recorder = call_recorder
        self._service_name = service_name
        self._name = name
        self._methods = methods
        self._url = url
        self._request_schema = request_schema
        self._response_schema = response_schema
        self._response_callback = response_callback

    @property
    def service_name(self):
        return self._service_name

    @property
    def name(self):
        return self._name

    @property
    def url(self):
        return self._url

    @property
    def methods(self):
        return list(self._methods)

    @property
    def call_recorder(self):
        return self._call_recorder

    def _validate(self, data_source_name, body, schema):
        if schema is not None:
            if body is None:
                error_list = ["Missing body"]
            else:
                if isinstance(body, bytes):
                    body = body.decode("utf-8")
                try:
                    data = json_loads(body)
                except ValueError as e :
                    error_list = ['JSON decoding error: {}'.format(e)]
                else:
                    error_list = validate(data, schema)
            if error_list:
                raise AssertionError(
                    VALIDATION_ERROR_TEXT.format(
                        data_source_name,
                        body,
                        schema,
                        self._service_name,
                        self._name,
                        self._url,
                        error_list,
                    )
                )

    def _validate_request(self, request):
        self._validate("request data", request.body, self._request_schema)

    def _validate_response(self, response_body):
        self._validate("response data", response_body, self._response_schema)

    def _record_response(
        self, request, response_status, response_headers, response_body
    ):
        # Shenanigans to get a response object like responses would
        # record in calls list
        def tmp_callback(request):
            return response_status, response_headers, response_body

        callback_response = responses.CallbackResponse(
            request.method, self._url, tmp_callback
        )
        response = callback_response.get_response(request)
        self._call_recorder.record(self, request, response, None)

    def _callback(self, request):
        try:
            self._validate_request(request)
            response_status, response_headers, response_body = self._response_callback(
                request
            )
            self._validate_response(response_body)
        except Exception as exc:
            self._call_recorder.record(self, request, None, exc)
            raise exc
        else:
            self._record_response(
                request, response_status, response_headers, response_body
            )
            return response_status, response_headers, response_body

    def get_calls(self):
        return self._call_recorder.get_calls_for(self)

    def get_last_call(self):
        return self.get_calls()[-1]

    def get_calls_matching(self, pattern):
        return self._call_recorder.get_calls_for_matching(self, pattern)

    def get_call_count(self):
        return len(self.get_calls())

    def was_called(self):
        return self.get_call_count() > 0


class EndpointMockContextManager(object):
    def __init__(
        self,
        methods,
        call_recorder,
        service_name,
        name,
        url,
        request_schema,
        response_schema,
        response_callback,
    ):
        self._methods = methods
        self._mock = EndpointMock(
            call_recorder,
            service_name,
            name,
            methods,
            url,
            request_schema,
            response_schema,
            response_callback,
        )

    def _start(self):
        responses_manager.attach_callback(
            self._methods, self._mock._url, self._mock._callback
        )

    def _stop(self):
        responses_manager.detach_callback(
            self._methods, self._mock._url, self._mock._callback
        )

    def __enter__(self):
        self._start()
        return self._mock

    def __exit__(self, *args):
        self._stop()


def response_callback_factory(status=200, headers=None, body=None, json=None):
    if headers is None:
        headers = CaseInsensitiveDict()
    else:
        headers = CaseInsensitiveDict(headers)
    if json is not None:
        assert body is None
        body = json_dumps(json).encode('utf-8')
        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'application/json'
    def response_callback(request):
        return status, headers, body
    return response_callback


ok_no_content_response_callback = response_callback_factory()


class Endpoint(object):
    """Configurable endpoint.

    Callable to create a context manager which activates and returns a mock
    for this endpoint.
    """
    def __init__(self, base_url, service_name, endpoint_spec, response_callback=None):
        self._url = urljoin(base_url, endpoint_spec.location)
        self._service_name = service_name
        self._name = endpoint_spec.name
        self._methods = list(endpoint_spec.methods)
        self._request_schema = endpoint_spec.request_schema
        self._response_schema = endpoint_spec.response_schema
        self._response_callback = response_callback
    
    @property
    def service_name(self):
        return self._service_name

    @property
    def name(self):
        return self._name

    @property
    def url(self):
        return self._url

    @property
    def methods(self):
        return list(self._methods)

    def disable_request_validation(self):
        self._request_schema = None

    def disable_response_validation(self):
        self._response_schema = None

    def disable_validation(self):
        self.disable_request_validation()
        self.disable_response_validation()

    def set_request_schema(self, schema):
        self._request_schema = schema

    def set_response_schema(self, schema):
        self._response_schema = schema

    def set_response_callback(self, callback):
        self._response_callback = callback
        
    def set_response(self, status=200, headers=None, body=None, json=None):
        self._response_callback = response_callback_factory(status, headers, body, json)

    def __call__(self, response_callback=None, call_recorder=None):
        if call_recorder is None:
            call_recorder = CallRecorder()
        if response_callback is None:
            response_callback = self._response_callback
            if response_callback is None:
                response_callback = ok_no_content_response_callback
        return EndpointMockContextManager(
            self._methods,
            call_recorder,
            self._service_name,
            self._name,
            self._url,
            self._request_schema,
            self._response_schema,
            response_callback,
        )


class ServiceMock(object):
    """Provides access to the endpoint mocks for this service and some functions 
    to get calls made to the services endpoints. 
     """
    def __init__(self, call_recorder, endpoints):
        self._call_recorder = call_recorder
        mocks = {}
        self._endpoint_context_managers = []
        for name, endpoint in endpoints.items():
            ecm = endpoint(call_recorder=call_recorder)
            mocks[name]= ecm._mock
            self._endpoint_context_managers.append(ecm)
        self.endpoints = Attrs(mocks)

    def get_calls(self):
        return self._call_recorder.get_calls()

    def get_calls_matching(self, pattern):
        return self._call_recorder.get_calls_matching(pattern)

    def get_call_count(self):
        return len(self.get_calls())

    def was_called(self):
        return self.get_call_count() > 0

    def _start(self):
        for ecm in self._endpoint_context_managers:
            ecm._start()

    def _stop(self):
        for ecm in self._endpoint_context_managers:
            ecm._stop()


class ServiceMockContextManager(object):
    def __init__(self, call_recorder, endpoints):
        self._mock = ServiceMock(call_recorder, endpoints)

    def __enter__(self):
        self._mock._start()
        return self._mock

    def __exit__(self, *args):
        self._mock._stop()


class Service(object):
    """Has configurable endpoints (.endpoints.*).

    Callable to create a context manager which will mock all the endpoints on
    the service.

    Endpoints can also be individually called to return a context manager 
    which just mocks that endpoint.
    """
    def __init__(self, base_url, name, endpoint_specs):
        self._base_url = base_url
        self._name = name
        endpoints = {}
        for endpoint_spec in endpoint_specs:
            endpoints[endpoint_spec.name] = Endpoint(self._base_url, self._name, endpoint_spec)
        self.endpoints = Attrs(endpoints)

    @property
    def name(self):
        return self._name

    @property
    def base_url(self):
        return self._base_url

    def __call__(self, call_recorder=None):
        if call_recorder is None:
            call_recorder = CallRecorder()
        return ServiceMockContextManager(call_recorder, dict(self.endpoints))


class ServiceFactory(object):
    """Callable to create Service instances.

    You can create multiple instances of a Service and configure each 
    independently.
    """
    def __init__(self, name, endpoint_specs):
        self._name = name
        self._endpoint_specs = endpoint_specs

    @property
    def name(self):
        return self._name

    def __call__(self, base_url):
        return Service(base_url, self.name, self._endpoint_specs)


__ALL__ = ['responses_mock_context', 'response_callback_factory', 'ServiceFactory', 'EndpointSpec', 'Endpoint']


# Copyright 2017-2018 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).

"""Build Service Doubles:

This module contains the entry point used to extract schemas from python source
files. The `main` function is installed as a console_script, and has several
modes of operation:

 - The 'scan_file' command allows a user to scan a random python source file
   and inspect what service doubles would be extracted from it. This is useful
   for ensuring that service_doubles can be extracted from a python source file
   before committing it.

 - The 'build' command takes a config file containing service names and
   locations, and builds a set of service_doubles based on that config.

In both cases, the service doubles are built by doing an AST parse of the
python source file in question and extracting calls to acceptable functions.
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import *  # NOQA

import argparse
import ast
import collections
import json
import logging
import os.path
import subprocess
import sys
import tempfile
import textwrap


try:
    FileNotFoundError
except NameError:
    PermissionError = FileNotFoundError = IOError


def main():
    args = parse_args()
    args.func(args)


def parse_args(arg_list=None, parser_class=None):
    parser = parser_class() if parser_class else argparse.ArgumentParser()
    subparser = parser.add_subparsers(dest='cmd')
    subparser.required = True
    scan_file_parser = subparser.add_parser(
        'scan-file', help='Scan a file, print extracted service doubles.')
    scan_file_parser.add_argument('file', type=str)
    scan_file_parser.set_defaults(func=scan_file)

    build_parser = subparser.add_parser(
        'build', help='build service doubles.')
    build_parser.add_argument('config_file', type=str)
    build_parser.set_defaults(func=build_service_doubles)

    return parser.parse_args(arg_list)


def scan_file(args):
    service_schemas = extract_schemas_from_file(args.file)
    print(render_service_double(
        'UNKNOWN', service_schemas, 'scan-file %s' % args.file))


def build_service_doubles(args):
    with tempfile.TemporaryDirectory() as workdir:
        service_config = read_service_config_file(args.config_file)
        target_root = os.path.dirname(args.config_file)
        for service_name in service_config['services']:
            service = service_config['services'][service_name]
            source_url = service['git_source']
            branch = service.get('git_branch')
            service_dir = fetch_service_source(
                workdir, service_name, source_url, branch)
            service_schemas = []
            for scan_path in service['scan_paths']:
                abs_path = os.path.join(service_dir, scan_path)
                service_schemas.extend(extract_schemas_from_file(abs_path))
            rendered = render_service_double(
                service_name, service_schemas, 'build %s' % args.config_file)
            write_service_double_file(target_root, service_name, rendered)
            print("Rendered schemas file for %s service: %d schemas" % (
                service, len(service_schemas)))


def read_service_config_file(config_path):
    with open(config_path, 'r') as config_file:
        return json.load(config_file)


def fetch_service_source(workdir, service_name, source_url, branch=None):
    print("Cloning source for %s service." % service_name)
    target_dir = os.path.join(workdir, service_name)
    cmd = ['git', 'clone']
    if branch is not None:
        cmd.extend(['-b', branch])
    cmd.extend([source_url, target_dir])
    subprocess.check_call(
        cmd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL
    )
    return target_dir


# ViewSchema contains all the information for a flask view...
ViewSchema = collections.namedtuple(
    'ViewSchema',
    [
        'view_name',      # The name of the view function.
        'version',        # The version the view was introduced at.
        'input_schema',   # The schema for requests to the service.
        'output_schema',  # The schema for responses from the service
        'methods',        # The methods this view supports.
        'url',            # The URL this view is mounted at.
        'doc',            # The documentation for this url
    ]
)


def extract_schemas_from_file(source_path):
    """Extract schemas from 'source_path'.

    :returns: a list of ViewSchema objects on success, None if no schemas
        could be extracted.
    """
    logging.info("Extracting schemas from %s", source_path)
    try:
        with open(source_path, 'r') as source_file:
            source = source_file.read()
    except (FileNotFoundError, PermissionError) as e:
        logging.error("Cannot extract schemas: %s", e.strerror)
    else:
        try:
            schemas = extract_schemas_from_source(source, source_path)
        except SyntaxError as e:
            logging.error("Cannot extract schemas: %s", str(e))
        else:
            logging.info(
                "Extracted %d %s",
                len(schemas),
                "schema" if len(schemas) == 1 else "schemas")
            return schemas


def _get_simple_assignments(tree):
    """Get simple assignments from node tree."""
    result = {}
    for node in ast.walk(tree):
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    result[target.id] = node.value
    return result


class _SimpleNamesResolver(ast.NodeTransformer):

    def __init__(self, names_values):
        super().__init__()
        self.names_values = names_values

    def visit_Name(self, node):
        if node.id in self.names_values:
            node = self.names_values[node.id]
        return node


def extract_schemas_from_source(source, filename='<unknown>'):
    """Extract schemas from 'source'.

    The 'source' parameter must be a string, and should be valid python
    source.

    If 'source' is not valid python source, a SyntaxError will be raised.

    :returns: a list of ViewSchema objects.
    """
    # Track which acceptable services have been configured.
    acceptable_services = set()
    # Track which acceptable views have been configured:
    acceptable_views = {}
    schemas_found = []
    ast_tree = ast.parse(source, filename)
    simple_names = _get_simple_assignments(ast_tree)

    assigns = [n for n in ast_tree.body if isinstance(n, ast.Assign)]
    call_assigns = [n for n in assigns if isinstance(n.value, ast.Call)]

    # We need to extract the AcceptableService-related views. We parse the
    # assignations twice: The first time to extract the AcceptableService
    # instances, the second to extract the views created on those services.
    for assign in call_assigns:
        if isinstance(assign.value.func, ast.Attribute):
            continue
        if assign.value.func.id == 'AcceptableService':
            for target in assign.targets:
                acceptable_services.add(target.id)

    for assign in call_assigns:
        # only consider calls which are attribute accesses, AND
        # calls where the object being accessed is in acceptable_services, AND
        # calls where the attribute being accessed is the 'api' method.
        if isinstance(assign.value.func, ast.Attribute) and \
           assign.value.func.value.id in acceptable_services and \
           assign.value.func.attr == 'api':
            # this is a view. We need to extract the url and methods specified.
            # they may be specified positionally or via a keyword.
            url = None
            name = None
            # methods has a default value:
            methods = ['GET']

            # This is a view - the URL is the first positional argument:
            args = assign.value.args
            if len(args) >= 1:
                url = ast.literal_eval(args[0])
            if len(args) >= 2:
                name = ast.literal_eval(args[1])
            kwargs = assign.value.keywords
            for kwarg in kwargs:
                if kwarg.arg == 'url':
                    url = ast.literal_eval(kwarg.value)
                if kwarg.arg == 'methods':
                    methods = ast.literal_eval(kwarg.value)
                if kwarg.arg == 'view_name':
                    name = ast.literal_eval(kwarg.value)
            if url and name:
                for target in assign.targets:
                    acceptable_views[target.id] = {
                        'url': url,
                        'name': name,
                        'methods': methods,
                    }

    # iterate over all functions, attempting to find the views.
    functions = [n for n in ast_tree.body if isinstance(n, ast.FunctionDef)]
    for function in functions:
        input_schema = None
        output_schema = None
        doc = ast.get_docstring(function)
        api_options_list = []
        for decorator in function.decorator_list:
            if not isinstance(decorator, ast.Call):
                continue
            if isinstance(decorator.func, ast.Attribute):
                decorator_name = decorator.func.value.id
                # extract version this view was introduced at, which can be
                # specified as an arg or a kwarg:
                version = None
                for kwarg in decorator.keywords:
                    if kwarg.arg == 'introduced_at':
                        version = ast.literal_eval(kwarg.value)
                        break
                if len(decorator.args) == 1:
                    version = ast.literal_eval(decorator.args[0])

                if decorator_name in acceptable_views:
                    api_options = acceptable_views[decorator_name]
                    api_options['version'] = version
                    api_options_list.append(api_options)
            else:
                decorator_name = decorator.func.id
                if decorator_name == 'validate_body':
                    _SimpleNamesResolver(simple_names).visit(decorator.args[0])
                    input_schema = ast.literal_eval(decorator.args[0])
                if decorator_name == 'validate_output':
                    _SimpleNamesResolver(simple_names).visit(decorator.args[0])
                    output_schema = ast.literal_eval(decorator.args[0])
        for api_options in api_options_list:
            schema = ViewSchema(
                    view_name=api_options['name'],
                    version=api_options['version'],
                    input_schema=input_schema,
                    output_schema=output_schema,
                    methods=api_options['methods'],
                    url=api_options['url'],
                    doc=doc,
                )
            schemas_found.append(schema)
    return schemas_found


def render_value(value):
    """Render a value, ensuring that any nested dicts are sorted by key."""
    if isinstance(value, list):
        return '[' + ', '.join(render_value(v) for v in value) + ']'
    elif isinstance(value, dict):
        return (
            '{' +
            ', '.join('{k!r}: {v}'.format(
                k=k, v=render_value(v)) for k, v in sorted(value.items())) +
            '}')
    else:
        return repr(value)


def render_service_double(service_name, schemas, regenerate_args):
    header = textwrap.dedent("""\
        # This file is AUTO GENERATED. Do not edit this file directly. Instead,
        # re-generate it by running '{progname} {regenerate_args}'.

        from acceptable._doubles import service_mock
        """.format(
            progname=os.path.basename(sys.argv[0]),
            regenerate_args=regenerate_args,
        ))

    rendered_schemas = []
    for schema in schemas:
        double_name = '%s_%s' % (
            schema.view_name, schema.version.replace('.', '_'))
        rendered_schema = textwrap.dedent("""\
        {double_name} = service_mock(
            service={service_name!r},
            methods={schema.methods!r},
            url={schema.url!r},
            input_schema={input_schema},
            output_schema={output_schema},
        )
        """).format(
            double_name=double_name,
            schema=schema,
            service_name=service_name,
            input_schema=render_value(schema.input_schema),
            output_schema=render_value(schema.output_schema),
        )

        rendered_schemas.append(rendered_schema)

    rendered_file = '{header}\n\n{schemas}\n'.format(
        header=header,
        schemas='\n\n'.join(rendered_schemas)
    )
    return rendered_file


def write_service_double_file(target_root, service_name, rendered):
    """Render syntactically valid python service double code."""
    target_path = os.path.join(
        target_root,
        'snapstore_schemas', 'service_doubles', '%s.py' % service_name
    )
    with open(target_path, 'w') as target_file:
        target_file.write(rendered)



# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import
from builtins import *  # NOQA
__metaclass__ = type

from future.utils import PY2, raise_from

import argparse
from collections import defaultdict, OrderedDict
from importlib import import_module
import json
import os
import sys

from jinja2 import (
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    PackageLoader,
)
import yaml

from acceptable import get_metadata, lint
from acceptable.dummy_importer import DummyImporterContext

if PY2:
    from future.types.newlist import newlist
    from future.types.newdict import newdict
    from yaml.representer import SafeRepresenter
    SafeRepresenter.add_representer(newlist, SafeRepresenter.represent_list)
    SafeRepresenter.add_representer(newdict, SafeRepresenter.represent_dict)


def tojson_filter(json_object, indent=4):
    return json.dumps(json_object, indent=indent)


TEMPLATES = Environment(
    loader=ChoiceLoader([
        FileSystemLoader('./'),
        PackageLoader('acceptable', 'templates'),
    ]),
    autoescape=False,
)
TEMPLATES.filters['tojson'] = tojson_filter


class ForceAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if not namespace.update:
            if namespace.metadata:
                namespace.metadata.close()  # suppresses resource warning
            parser.error('--force can only be used with --update')
        else:
            namespace.force = True


def main():
    try:
        cli_args = parse_args()
        sys.exit(cli_args.func(cli_args))
    except Exception as e:
        raise
        sys.exit(str(e))


def parse_args(raw_args=None, parser_cls=None, stdin=None, stdout=None):
    if parser_cls is None:
        parser_cls = argparse.ArgumentParser
    if stdin is None:
        stdin = sys.stdin
    if stdout is None:
        stdout = sys.stdout

    parser = parser_cls(
        description='Tool for working with acceptable metadata',
    )
    subparser = parser.add_subparsers(dest='cmd')
    subparser.required = True

    metadata_parser = subparser.add_parser(
        'metadata', help='Import project and print extracted metadata in JSON')
    metadata_parser.add_argument('modules', nargs='+')
    metadata_parser.add_argument(
        '--dummy-dependencies', '-d',
        action='store_true',
        default=False,
        help='Import code in a sandbox where dependencies are mocked',
    )
    metadata_parser.add_argument(
        '--output',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout,
        help='metadata output file path, uses stdout if omitted'
    )
    metadata_parser.set_defaults(func=metadata_cmd)

    render_parser = subparser.add_parser(
        'render', help='Render markdown documentation for API metadata'
    )
    render_parser.add_argument(
        'metadata',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin,
    )
    render_parser.add_argument(
        '--name', '-n', required=True, help='Name of service')
    render_parser.add_argument(
        '--dir', '-d', default='docs', help='output directory')
    render_parser.add_argument(
        '--page-template',
        type=TEMPLATES.get_template,
        default=TEMPLATES.get_template('api_group.md.j2'),
        help='Jinja2 template to render each API',
    )
    render_parser.add_argument(
        '--index-template',
        type=TEMPLATES.get_template,
        default=TEMPLATES.get_template('index.md.j2'),
        help='Jinja2 template to render the API index page',
    )
    render_parser.add_argument(
        '--extension', '-e',
        default='md',
        help='File extension for rendered documentation',
    )

    render_parser.set_defaults(func=render_cmd)

    lint_parser = subparser.add_parser(
        'lint', help='Compare current metadata against file metadata')
    lint_parser.add_argument(
        'metadata',
        nargs='?',
        type=argparse.FileType('r'),
    )
    lint_parser.add_argument('modules', nargs='+')
    lint_parser.add_argument(
        '-q', '--quiet',
        action='store_true',
        default=False,
        help='Do not emit warnings',
    )
    lint_parser.add_argument(
        '--strict', '--pedantic', '--overhead',
        action='store_true',
        default=False,
        help='Even warnings count as failure',
    )
    lint_parser.add_argument(
        '--update',
        action='store_true',
        default=False,
        help='Update metadata file to new metadata if lint passes',
    )
    lint_parser.add_argument(
        '--force',
        action=ForceAction,
        nargs=0,
        default=False,
        help='Update metadata even if linting fails',
    )

    lint_parser.set_defaults(func=lint_cmd)

    doubles_parser = subparser.add_parser(
        'doubles',
        help='Generate test doubles'
    )
    doubles_parser.add_argument(
        'metadata',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin,
        help='metadata file path, uses stdin if omitted'
    )
    doubles_parser.add_argument(
        '-n', '--new-style',
        action='store_true',
        default=False,
        help='Generate new style ServiceFactory mocks',
    )
    doubles_parser.set_defaults(func=doubles_cmd)

    version_parser = subparser.add_parser(
        'api-version',
        help='Get the current API version from JSON meta, and '
             'optionally from current code also',
    )
    version_parser.add_argument(
        'metadata',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin,
        help='The JSON metadata for the API',
    )
    version_parser.add_argument(
        'modules',
        nargs='*',
        help='Optional modules to import for current imported API',
    )
    version_parser.set_defaults(func=version_cmd)

    return parser.parse_args(raw_args)


def metadata_cmd(cli_args):
    import_metadata(cli_args.modules, cli_args.dummy_dependencies)
    current, _ = get_metadata().serialize()
    cli_args.output.write(json.dumps(current, indent=2))


def add_working_dir_to_python_path():
    cwd = os.getcwd()
    if cwd not in sys.path:
        sys.path.insert(0, cwd)


def import_or_exec(path):
    """If path exists and ends in .py it is exec'd otherwise
    import is attempted"""
    if os.path.exists(path) and path.endswith('.py'):
        try:
            exec(open(path).read(), {}, {})
        except:
            raise Exception('Could not exec {!r}'.format(path))
    else:
        try:
            import_module(path)
        except:
            raise Exception(
                '{!r} did not look like a filepath and could not be '
                'loaded as a module'.format(path)
            )


def import_metadata_dummy_dependencies(module_paths):
    import acceptable._service
    add_working_dir_to_python_path()
    for path in module_paths:
        with DummyImporterContext(path):
            import_or_exec(path)


def import_metadata_real_dependencies(module_paths):
    add_working_dir_to_python_path()
    for path in module_paths:
        import_or_exec(path)


def import_metadata(module_paths, dummy_dependencies=False):
    """Imports modules or execs filepaths in order
    to get acceptable decorator metadata.
    """
    if dummy_dependencies:
        import_metadata_dummy_dependencies(module_paths)
    else:
        import_metadata_real_dependencies(module_paths)


def load_metadata(stream):
    """Load JSON metadata from opened stream."""
    try:
        metadata = json.load(
            stream, encoding='utf8', object_pairs_hook=OrderedDict)
    except json.JSONDecodeError as e:
        err = RuntimeError('Error parsing {}: {}'.format(stream.name, e))
        raise_from(err, e)
    else:
        # convert changelog keys back to ints for sorting
        for group in metadata:
            if group == '$version':
                continue
            apis = metadata[group]['apis']
            for api in apis.values():
                int_changelog = OrderedDict()
                for version, log in api.get('changelog', {}).items():
                    int_changelog[int(version)] = log
                api['changelog'] = int_changelog
    finally:
        stream.close()

    return metadata


def render_cmd(cli_args):
    root_dir = cli_args.dir
    en_dir = os.path.join(root_dir, 'en')
    if not os.path.exists(en_dir):
        os.makedirs(en_dir)
    metadata = load_metadata(cli_args.metadata)

    for path, content in render_markdown(metadata, cli_args):
        full_path = os.path.join(root_dir, path)
        with open(full_path, 'w', encoding='utf8') as f:
            f.write(content)


def render_markdown(metadata, cli_args):
    navigation = [{
        'title': 'Index',
        'location': 'index.' + cli_args.extension,
    }]
    version = metadata.pop('$version', None)
    changelog = defaultdict(dict)

    for group in metadata:
        apis = metadata[group]['apis']
        for api in apis.values():
            # collect global changelog
            for changed_version, log in api.get('changelog', {}).items():
                changelog[changed_version][(group, api['api_name'])] = log

        documented_apis = [
            api for api in apis.values()
            if not api.get('undocumented', False)
        ]

        if documented_apis:
            group_apis = []
            deprecated_apis = []
            for api in documented_apis:
                if api.get('deprecated_at', False):
                    deprecated_apis.append(api)
                else:
                    group_apis.append(api)
            sorted_apis = group_apis + deprecated_apis

            page_file = '{}.{}'.format(group, cli_args.extension)
            page = {'title': group.title(), 'location': page_file}
            navigation.append(page)

            path = os.path.join('en', page_file)
            yield path, cli_args.page_template.render(
                group_name=group,
                group_title=metadata[group].get('title', group),
                group_apis=sorted_apis,
                group_doc=metadata[group].get('docs', ''),
            )

    yield (
        os.path.join('en', 'index.' + cli_args.extension),
        cli_args.index_template.render(
            version=version,
            service_name=cli_args.name,
            changelog=changelog,
        )
    )

    # documentation-builder requires yaml metadata files in certain locations
    yield os.path.join('en', 'metadata.yaml'), yaml.safe_dump(
        {'navigation': navigation},
        default_flow_style=False,
        encoding=None,
    )
    site_meta = {
        'site_title': '{} Documentation: version {}'.format(
            cli_args.name,
            version,
        )
    }
    yield 'metadata.yaml', yaml.safe_dump(
        site_meta,
        default_flow_style=False,
        encoding=None,
    )


def lint_cmd(cli_args, stream=sys.stdout):
    metadata = load_metadata(cli_args.metadata)
    import_metadata(cli_args.modules)
    current, locations = get_metadata().serialize()

    has_errors = False
    display_level = lint.WARNING
    error_level = lint.DOCUMENTATION

    if cli_args.strict:
        display_level = lint.WARNING
        error_level = lint.WARNING
    elif cli_args.quiet:
        display_level = lint.DOCUMENTATION

    for message in lint.metadata_lint(metadata, current, locations):
        if message.level >= display_level:
            stream.write('{}\n'.format(message))

        if message.level >= error_level:
            has_errors = True

    if cli_args.update:
        if not has_errors or cli_args.force:
            with open(cli_args.metadata.name, 'w') as f:
                json.dump(current, f, indent=2)

    return 1 if has_errors else 0


def doubles_cmd(cli_args, stream=sys.stdout):
    metadata = json.load(cli_args.metadata, encoding='utf8')
    if cli_args.new_style:
        from . import generate_mocks
        generate_mocks.generate_service_factory(
            metadata,
            stream=stream,
        )
    else:
        from . import generate_doubles
        generate_doubles.generate_service_mock_doubles(
            metadata,
            stream=stream
        )


def version_cmd(cli_args, stream=sys.stdout):
    metadata = load_metadata(cli_args.metadata)
    json_version = metadata['$version']
    import_version = None

    if cli_args.modules:
        import_metadata(cli_args.modules)
        import_version = get_metadata().current_version

    stream.write('{}: {}\n'.format(cli_args.metadata.name, json_version))

    if import_version is not None:
        if len(cli_args.modules) == 1:
            name = cli_args.modules[0]
        else:
            name = 'Imported API'
        stream.write('{}: {}\n'.format(name, import_version))

    return 0


if __name__ == '__main__':
    main()



# Copyright 2019 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).
from __future__ import absolute_import

import responses


class ResponsesManager(object):
    """The responses library is used to add mock behaviour into the requests
    library. 
    
    It does this using the RequestsMock class, however only one of 
    these can be active at a time. Attempting to start a new RequestsMock
    will remove any others hooked into requests.
    
    We use an instance of this class to manage use of the RequestsMock 
    instance `responses.mock`. This allows us to start, stop and reset 
    the it at the right time.
    """
    def __init__(self):
        self._attached = 0

    def attach(self):
        if self._attached == 0:
            responses.mock.start()
        self._attached += 1

    def detach(self):
        self._attached -= 1
        assert self._attached >= 0
        if self._attached == 0:
            responses.mock.stop()
            responses.mock.reset()

    def attach_callback(self, methods, url, callback):
        for method in methods:
            responses.mock.add_callback(method, url, callback)
        self.attach()

    def detach_callback(self, methods, url, callback):
        for method in methods:
            responses.mock.remove(method, url)
        self.detach()


responses_manager = ResponsesManager()


class responses_mock_context(object):
    """Provides access to `responses.mock` in a way that is safe to mix with
    mocks from `acceptable.mocks`.

    Use as a context manager or a decorator:

        def blah():
            with responese_mock_context():
                ...
    Or:

        @responese_mock_context()
        def blah():
            ,,,
    """
    def __enter__(self):
        responses_manager.attach()
        return responses.mock

    def __exit__(self, *args):
        responses_manager.detach()

    def __call__(self, func):
        # responses.get_wrapper creates a function which has the same
        # signature etc.  as `func`. It execs `wrapper_template`
        # in a seperate namespace to do this. See get_wrapper code.
        wrapper_template = """\
def wrapper%(signature)s:
    with responses_mock_context:
        return func%(funcargs)s
"""
        namespace = {'responses_mock_context': self, 'func': func}
        return responses.get_wrapped(func, wrapper_template, namespace)



# Copyright 2019 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).
from __future__ import print_function
import json
import textwrap
from sys import stdout


def write_double(stream, api_name, api_data):
    double_name = u'%s_%s' % (api_name, "1_0")
    double_text = textwrap.dedent(u"""\

    {double_name} = service_mock(
        service={service!r},
        methods={methods!r},
        url={url!r},
        input_schema={request_schema!r},
        output_schema={response_schema!r},
    )

    """).format(
        double_name=double_name,
        **api_data
    )
    stream.write(double_text)


HEADER_TEXT = u"""\
# This file is AUTO GENERATED. Do not edit this file directly. Instead,
# re-generate it.

from acceptable._doubles import service_mock

"""


def write_header(stream):
    stream.write(HEADER_TEXT)


def generate_service_mock_doubles(metadata, stream=stdout):
    write_header(stream)
    for module_name, module_data in metadata.items():
        if module_name.startswith('$'):
            continue
        for api_name, api_data in module_data['apis'].items():
            write_double(stream, api_name, api_data)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import *  # NOQA

from collections import OrderedDict
import difflib
import inspect
import pprint
import sys
import textwrap


def get_callsite_location(depth=1):
    frame = sys._getframe(depth + 1)
    return {
        'filename': inspect.getsourcefile(frame),
        'lineno': frame.f_lineno,
        'module': inspect.getmodule(frame),
    }


def clean_docstring(docstring):
    """Dedent docstring, special casing the first line."""
    docstring = docstring.strip()
    if '\n' in docstring:
        # multiline docstring
        if docstring[0].isspace():
            # whole docstring is indented
            return textwrap.dedent(docstring)
        else:
            # first line not indented, rest maybe
            first, _, rest = docstring.partition('\n')
            return first + '\n' + textwrap.dedent(rest)
    return docstring


def _sort_schema(schema):
    """Recursively sorts a JSON schema by dict key."""

    if isinstance(schema, dict):
        for k, v in sorted(schema.items()):
            if isinstance(v, dict):
                yield k, OrderedDict(_sort_schema(v))
            elif isinstance(v, list):
                yield k, list(_sort_schema(v))
            else:
                yield k, v
    elif isinstance(schema, list):
        for v in schema:
            if isinstance(v, dict):
                yield OrderedDict(_sort_schema(v))
            elif isinstance(v, list):
                yield list(_sort_schema(v))
            else:
                yield v
    else:
        yield d


def sort_schema(schema):
    sorted_schema = OrderedDict(_sort_schema(schema))
    # ensure sorting the schema does not alter it
    if schema != sorted_schema:
        d1 = pprint.pformat(schema).splitlines()
        d2 = pprint.pformat(sorted_schema).splitlines()
        diff = '\n'.join(difflib.ndiff(d1, d2))
        raise RuntimeError('acceptable: sorting schema failed:\n' + diff)
    return sorted_schema




# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).

"""acceptable - Programatic API Metadata for Flask apps."""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import *  # NOQA
__metaclass__ = type

import argparse
import json
import sys

from django.core.management.base import BaseCommand, CommandParser

from acceptable import get_metadata
from acceptable.djangoutil import get_urlmap
from acceptable.__main__ import load_metadata


class Command(BaseCommand):
    help = 'Generate Acceptable API Metadata from project'

    def add_arguments(self, parser):
        cmd = self

        class SubParser(CommandParser):
            """Django command aware subparser."""
            def __init__(self, **kwargs):
                super().__init__(cmd, **kwargs)

        subparser = parser.add_subparsers(dest='cmd', parser_class=SubParser)
        subparser.required = True

        metadata_parser = subparser.add_parser(
            'metadata',
            help='Import project and print extracted metadata in json')
        metadata_parser.set_defaults(func=self.metadata)

        version_parser = subparser.add_parser(
            'api-version',
            help='Get the current api version from json meta, and '
                 'optionally from current code also',
        )
        version_parser.add_argument(
            'metadata',
            nargs='?',
            type=argparse.FileType('r'),
            default=sys.stdin,
            help='The json metadata for the api',
        )
        version_parser.set_defaults(func=self.version)

    def handle(self, *args, **options):
        get_urlmap()  # this imports all urls and initialises the url mappings
        func = options['func']
        current, _ = get_metadata().serialize()
        func(options, current)

    def metadata(self, options, current):
        print(json.dumps(current, indent=2))

    def version(self, options, current, stream=sys.stdout):
        file_metadata = load_metadata(options['metadata'])
        json_version = file_metadata['$version']
        import_version = current['$version']
        stream.write('{}: {}\n'.format(options['metadata'].name, json_version))
        stream.write('Imported API: {}\n'.format(import_version))



# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import *  # NOQA
__metaclass__ = type

import logging

import django
from django.forms import widgets, fields

from acceptable._service import AcceptableAPI
from acceptable.util import clean_docstring, sort_schema

logger = logging.getLogger('acceptable')
_urlmap = None

if django.VERSION >= (2, 0):
    from django.urls import (
        URLPattern,
        URLResolver,
        get_resolver,
    )
    PATTERNS = (URLPattern,)
    RESOLVERS = (URLResolver,)

    def get_pattern(p):
        return str(p.pattern)
else:
    try:
        from django.urls import (
            LocaleRegexURLResolver,
            RegexURLPattern,
            RegexURLResolver,
            get_resolver,
        )
    except ImportError:
        from django.core.urlresolvers import (
            LocaleRegexURLResolver,
            RegexURLPattern,
            RegexURLResolver,
            get_resolver,
        )

    PATTERNS = (RegexURLPattern,)
    RESOLVERS = (RegexURLResolver, LocaleRegexURLResolver)

    def get_pattern(p):
        return p.regex.pattern


def get_urlmap():
    global _urlmap
    if _urlmap is None:
        resolver = get_resolver()
        _urlmap = dict(urlmap(resolver.url_patterns))
    return _urlmap


def urlmap(patterns):
    """Recursively build a map of (group, name) => url patterns.

    Group is either the resolver namespace or app name for the url config.

    The urls are joined with any prefixes, and cleaned up of extraneous regex
    specific syntax."""
    for pattern in patterns:
        group = getattr(pattern, 'namespace', None)
        if group is None:
            group = getattr(pattern, 'app_name', None)
        path = '/' + get_pattern(pattern).lstrip('^').rstrip('$')
        if isinstance(pattern, PATTERNS):
            yield (group, pattern.name), path
        elif isinstance(pattern, RESOLVERS):
            subpatterns = pattern.url_patterns
            for (_, name), subpath in urlmap(subpatterns):
                yield (group, name), path.rstrip('/') + subpath


def get_field_schema(name, field):
    """Returns a JSON Schema representation of a form field."""
    field_schema = {
        'type': 'string',
    }

    if field.label:
        field_schema['title'] = str(field.label)  # force translation

    if field.help_text:
        field_schema['description'] = str(field.help_text)  # force translation

    if isinstance(field, (fields.URLField, fields.FileField)):
        field_schema['format'] = 'uri'
    elif isinstance(field, fields.EmailField):
        field_schema['format'] = 'email'
    elif isinstance(field, fields.DateTimeField):
        field_schema['format'] = 'date-time'
    elif isinstance(field, fields.DateField):
        field_schema['format'] = 'date'
    elif isinstance(field, (fields.DecimalField, fields.FloatField)):
        field_schema['type'] = 'number'
    elif isinstance(field, fields.IntegerField):
        field_schema['type'] = 'integer'
    elif isinstance(field, fields.NullBooleanField):
        field_schema['type'] = 'boolean'
    elif isinstance(field.widget, widgets.CheckboxInput):
        field_schema['type'] = 'boolean'

    if getattr(field, 'choices', []):
        field_schema['enum'] = sorted([choice[0] for choice in field.choices])

    # check for multiple values
    if isinstance(field.widget, (widgets.Select, widgets.ChoiceWidget)):
        if field.widget.allow_multiple_selected:
            # promote to array of <type>, move details into the items field
            field_schema['items'] = {
                'type': field_schema['type'],
            }
            if 'enum' in field_schema:
                field_schema['items']['enum'] = field_schema.pop('enum')
            field_schema['type'] = 'array'

    return field_schema


def get_form_schema(form):
    """Return a JSON Schema object for a Django Form."""
    schema = {
        'type': 'object',
        'properties': {},
    }

    for name, field in form.base_fields.items():
        schema['properties'][name] = get_field_schema(name, field)
        if field.required:
            schema.setdefault('required', []).append(name)

    return schema


class DjangoAPI(AcceptableAPI):
    """Django-flavour API metadata

    Supports setting a Django form to provide json schema for documentation, as
    well providing an API handler class to inspect for more metadata.
    """

    def __init__(
            self,
            service,
            name,
            introduced_at,
            options={},
            location=None,
            undocumented=False,
            deprecated_at=None,
            title=None):
        # leave url blank, as we can't know it until django has set itself up
        # properly
        super().__init__(
            service,
            name,
            None,
            introduced_at,
            options,
            location,
            undocumented,
            deprecated_at,
            title,
        )

        self._form = None
        self.handler_class = None

    def resolve_url(self):
        name = self.name
        try_default = True

        if ':' in self.name:
            group, name = self.name.split(':', 2)
            try_default = False  # user passed explicit group, just use that
        else:
            group = self.service.group

        urlmap = get_urlmap()
        url = urlmap.get((group, name))
        if url is None and try_default:
            url = urlmap.get((None, name))
        # TODO should we error out if there is no map with that name?
        return url

    @property
    def methods(self):
        default = ['GET']
        if 'methods' in self.options:
            return list(self.options.get('methods', default))

        # allowed_methods works for piston handlers
        # TODO: add support for DRF? And maybe plain view functions with
        # decorators?
        return list(getattr(self.handler_class, 'allowed_methods', default))

    @property
    def django_form(self):
        return self._form

    @django_form.setter
    def django_form(self, form):
        self._form = form
        schema = get_form_schema(form)
        self.request_schema = sort_schema(schema)

    def handler(self, handler_class):
        """Link to an API handler class (e.g. piston or DRF)."""
        self.handler_class = handler_class
        # we take the docstring from the handler class, not the methods
        if self.docs is None and handler_class.__doc__:
            self.docs = clean_docstring(handler_class.__doc__)
        return handler_class



# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).

"""acceptable - Programatic API Metadata for Flask apps."""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import *  # NOQA
__metaclass__ = type

from collections import OrderedDict, defaultdict
import textwrap

from acceptable import _validation
from acceptable.util import (
    clean_docstring,
    get_callsite_location,
    sort_schema,
)


class InvalidAPI(Exception):
    pass


class APIMetadata():
    """Global datastructure for all services.

    Provides a single point to register apis against, so we can easily inspect
    and verify uniqueness.
    """

    def __init__(self):
        self.services = OrderedDict()
        self.api_names = set()
        self.urls = set()
        self._current_version = None

    def register_service(self, service, group, docs=None, title=None):
        if service not in self.services:
            self.services[service] = OrderedDict()

        if group not in self.services[service]:
            self.services[service][group] = APIGroup(group, docs, title)
        elif docs is not None:
            additional_docs = '\n' + clean_docstring(docs)
            self.services[service][group].docs += additional_docs

        return self.services[service][group]

    def register_api(self, service, group, api):
        # check for name/url clashes globally
        # FIXME: this should probably be per service?
        if api.name in self.api_names:
            raise InvalidAPI(
                'API {} is already registered in service {}'.format(
                    api.name, service)
            )
        self.api_names.add(api.name)

        if api.url is not None:
            url_key = (api.url, tuple(api.methods))
            if url_key in self.urls:
                raise InvalidAPI(
                    'URL {} {} is already in service {}'.format(
                        '|'.join(api.methods), api.url, service)
                )
            self.urls.add(url_key)

        self.services[service][group][api.name] = api

    @property
    def current_version(self):
        if self._current_version is None:
            versions = set()
            for service in self.services.values():
                for group in service.values():
                    for api in group.values():
                        versions.add(api.introduced_at)
                        if api._changelog:
                            versions.add(max(api._changelog))
            if versions:
                self._current_version = max(versions)
        return self._current_version

    def bind(self, flask_app, service, group=None):
        """Bind the service API urls to a flask app."""
        if group not in self.services[service]:
            raise RuntimeError(
                'API group {} does not exist in service {}'.format(
                    group, service)
            )
        for name, api in self.services[service][group].items():
            # only bind APIs that have views associated with them
            if api.view_fn is None:
                continue
            if name not in flask_app.view_functions:
                flask_app.add_url_rule(
                    api.url, name, view_func=api.view_fn, **api.options)

    def bind_all(self, flask_app):
        for service, groups in self.services.items():
            for group in groups:
                self.bind(flask_app, service, group)

    def clear(self):
        self.services.clear()
        self.api_names.clear()
        self.urls.clear()
        self._current_version = None

    def groups(self):
        for service, groups in self.services.items():
            for group in groups.values():
                yield service, group

    def serialize(self):
        """Serialize into JSONable dict, and associated locations data."""
        api_metadata = OrderedDict()
        # $ char makes this come first in sort ordering
        api_metadata['$version'] = self.current_version
        locations = {}

        for svc_name, group in self.groups():
            group_apis = OrderedDict()
            group_metadata = OrderedDict()
            group_metadata['apis'] = group_apis
            group_metadata['title'] = group.title
            api_metadata[group.name] = group_metadata

            if group.docs is not None:
                group_metadata['docs'] = group.docs

            for name, api in group.items():
                group_apis[name] = OrderedDict()
                group_apis[name]['service'] = svc_name
                group_apis[name]['api_group'] = group.name
                group_apis[name]['api_name'] = api.name
                group_apis[name]['introduced_at'] = api.introduced_at
                group_apis[name]['methods'] = api.methods
                group_apis[name]['request_schema'] = api.request_schema
                group_apis[name]['response_schema'] = api.response_schema
                group_apis[name]['doc'] = api.docs
                group_apis[name]['changelog'] = api._changelog
                if api.title:
                    group_apis[name]['title'] = api.title
                else:
                    title = name.replace('-', ' ').replace('_', ' ').title()
                    group_apis[name]['title'] = title

                group_apis[name]['url'] = api.resolve_url()

                if api.undocumented:
                    group_apis[name]['undocumented'] = True
                if api.deprecated_at is not None:
                    group_apis[name]['deprecated_at'] = api.deprecated_at

                locations[name] = {
                    'api': api.location,
                    'request_schema': api._request_schema_location,
                    'response_schema': api._response_schema_location,
                    'changelog': api._changelog_locations,
                    'view': api.view_fn_location,
                }

        return api_metadata, locations


_metadata = None


def get_metadata():
    global _metadata
    if _metadata is None:
        _metadata = APIMetadata()
    return _metadata


def clear_metadata():
    global _metadata
    _metadata = None


class APIGroup(OrderedDict):
    """Wrapper for collection of APIs, with associated documentation."""
    def __init__(self, name=None, docs=None, title=None):
        self.name = name
        self.title = title
        if self.name is None:
            self.name = 'default'
            self.title = 'Default'
        elif title is None:
            self.title = name.replace('-', ' ').title()
        self.docs = docs
        super().__init__()


class AcceptableService():
    """User facing API for a service using acceptable to manage API versions.

    This provides a nicer interface to manage the global API metadata within
    a single file.

    It is just a factory and proxy to the global metadata state, it does not
    store any API state internally.
    """

    def __init__(self, name, group=None, title=None, metadata=None):
        """Create an instance of AcceptableService.

        :param name: The service name.
        :param group: An arbitrary API group within a service.
        """
        self.name = name
        self.group = group
        if metadata is None:
            self.metadata = get_metadata()
        else:
            self.metadata = metadata

        self.location = get_callsite_location()
        self.doc = None
        module = self.location['module']
        docs = None
        if module and module.__doc__:
            docs = clean_docstring(module.__doc__)
        self.metadata.register_service(name, group, docs, title)

    @property
    def apis(self):
        return self.metadata.services[self.name][self.group]

    def api(self,
            url,
            name,
            introduced_at=None,
            undocumented=False,
            deprecated_at=None,
            title=None,
            **options):
        """Add an API to the service.

        :param url: This is the url that the API should be registered at.
        :param name: This is the name of the api, and will be registered with
            flask apps under.

        Other keyword arguments may be used, and they will be passed to the
        flask application when initialised. Of particular interest is the
        'methods' keyword argument, which can be used to specify the HTTP
        method the URL will be added for.
        """
        location = get_callsite_location()
        api = AcceptableAPI(
            self,
            name,
            url,
            introduced_at,
            options,
            undocumented=undocumented,
            deprecated_at=deprecated_at,
            title=title,
            location=location,
        )
        self.metadata.register_api(self.name, self.group, api)
        return api

    def django_api(
            self,
            name,
            introduced_at,
            undocumented=False,
            deprecated_at=None,
            title=None,
            **options):
        """Add a django API handler to the service.

        :param name: This is the name of the django url to use.

        The 'methods' paramater can be supplied as normal, you can also user
        the @api.handler decorator to link this API to its handler.

        """
        from acceptable.djangoutil import DjangoAPI
        location = get_callsite_location()
        api = DjangoAPI(
            self,
            name,
            introduced_at,
            options,
            location=location,
            undocumented=undocumented,
            deprecated_at=deprecated_at,
            title=title,
        )
        self.metadata.register_api(self.name, self.group, api)
        return api

    def bind(self, flask_app):
        """Bind the service API urls to a flask app."""
        self.metadata.bind(flask_app, self.name, self.group)

    # b/w compat
    initialise = bind


class AcceptableAPI():
    """Metadata abount an api endpoint."""

    def __init__(
            self,
            service,
            name,
            url,
            introduced_at,
            options={},
            location=None,
            undocumented=False,
            deprecated_at=None,
            title=None):

        self.service = service
        self.name = name
        self.url = url
        self.introduced_at = introduced_at
        self.options = options
        self.view_fn = None
        self.view_fn_location = None
        self.docs = None
        self._request_schema = None
        self._request_schema_location = None
        self._response_schema = None
        self._response_schema_location = None
        self._changelog = OrderedDict()
        self._changelog_locations = OrderedDict()
        if location is None:
            self.location = get_callsite_location()
        else:
            self.location = location
        self.undocumented = undocumented
        self.deprecated_at = deprecated_at
        self.title = title

    @property
    def methods(self):
        return list(self.options.get('methods', ['GET']))

    def resolve_url(self):
        return self.url

    @property
    def request_schema(self):
        return self._request_schema

    @request_schema.setter
    def request_schema(self, schema):
        if schema is not None:
            _validation.validate_schema(schema)
        self._request_schema = sort_schema(schema)
        # this location is the last item in the dict, sadly
        self._request_schema_location = get_callsite_location()

    @property
    def response_schema(self):
        return self._response_schema

    @response_schema.setter
    def response_schema(self, schema):
        if schema is not None:
            _validation.validate_schema(schema)
        self._response_schema = sort_schema(schema)
        # this location is the last item in the dict, sadly
        self._response_schema_location = get_callsite_location()

    def changelog(self, api_version, doc):
        """Add a changelog entry for this api."""
        doc = textwrap.dedent(doc).strip()
        self._changelog[api_version] = doc
        self._changelog_locations[api_version] = get_callsite_location()

    def __call__(self, fn):
        wrapped = fn
        if self.response_schema:
            wrapped = _validation.wrap_response(wrapped, self.response_schema)
        if self.request_schema:
            wrapped = _validation.wrap_request(wrapped, self.request_schema)

        location = get_callsite_location()
        # this will be the lineno of the last decorator, so we want one
        # below it for the actual function
        location['lineno'] += 1
        self.register_view(wrapped, location)
        return wrapped

    def register_view(self, view_fn, location=None, introduced_at=None):
        if self.view_fn is not None:
            raise InvalidAPI('api already has view registered')
        self.view_fn = view_fn
        self.view_fn_location = location
        if self.introduced_at is None:
            self.introduced_at = introduced_at
        if self.docs is None and self.view_fn.__doc__ is not None:
            self.docs = clean_docstring(self.view_fn.__doc__)

    # legacy view decorator
    def view(self, introduced_at):

        def decorator(fn):
            location = get_callsite_location()
            # this will be the lineno of the last decorator, so we want one
            # below it for the actual function
            location['lineno'] += 1

            # convert older style version strings
            if introduced_at == '1.0':
                self.introduced_at = 1
            elif introduced_at is not None:
                self.introduced_at = int(introduced_at)

            self.register_view(fn, location, introduced_at)

            # support for legacy @validate_{body,output} decorators
            # we don't know the order of decorators, so allow for both.
            # Note that if these schemas come from the @validate decorators,
            # they are already validated, so we set directly.
            fn._acceptable_metadata = self
            if self._request_schema is None:
                self._request_schema = getattr(fn, '_request_schema', None)
                self._request_schema_location = getattr(
                    fn, '_request_schema_location', None)
            if self._response_schema is None:
                self._response_schema = getattr(fn, '_response_schema', None)
                self._response_schema_location = getattr(
                    fn, '_response_schema_location', None)

            return fn

        return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2017 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).

from setuptools import find_packages, setup


VERSION = '0.29'


setup(
    name='acceptable',
    version=VERSION,
    description='API version negotiation for flask-based web services.',
    author='Canonical Online Services',
    author_email='online-services@lists.canonical.com',
    url='https://github.com/canonical-ols/acceptable',
    license='LGPLv3',
    packages=find_packages(exclude=['examples', '*tests']),
    long_description=''.join(open('README.rst').readlines()[2:]),
    long_description_content_type='text/x-rst',
    install_requires=[
        'future',
        'jinja2',
        'jsonschema',
        'pyyaml',
    ],
    extras_require=dict(
        flask=[
            'Flask<2.0',
        ],
        django=[
            'django>=1.11,<2.1',
        ]
    ),
    test_suite='acceptable.tests',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'build_service_doubles = acceptable._build_doubles:main',
            'acceptable = acceptable.__main__:main',
        ]
    },
)



# Copyright 2019 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).
from __future__ import print_function
from future import standard_library

standard_library.install_aliases()
import sys
try:
    # install_aliases() doesn't seem to include unittest.mock
    from mock import MagicMock
except ImportError:
    from unittest.mock import MagicMock


class DummyFinder(object):
    """Implements PEP 302 module finder and loader which will pretend to
    load modules but actually just create mocks that look like them.

    This allows python code to be loaded when its dependencies are not
    installed.

    allowed_real_modules is a list of module names to not mock instead the
    module finders are tried.

    finders can come from sys.meta_path

    modules in passthrough will raise the correct error on import if
    they can't be loaded.

    This class is used by DummyImporterContext which patches sys.modules
    and sys.meta_path.
    """

    def __init__(self, allowed_real_modules, finders):
        self.finders = finders
        self.allowed = set(allowed_real_modules)

    def find_module(self, fullname, path=None):
        if fullname in self.allowed:
            for finder in self.finders:
                loader = finder.find_module(fullname, path)
                if loader is not None:
                    return loader
            else:
                return None
        return self

    def load_module(self, fullname):
        try:
            return sys.modules[fullname]
        except KeyError:
            pass
        mod = MagicMock()
        sys.modules[fullname] = mod
        mod.__file__ = "<DummyFinder>"
        mod.__loader__ = self
        mod.__path__ = []
        mod.__package__ = fullname
        mod.__doc__ = "DummyImporterContext dummy"
        return mod


class DummyImporterContext(object):
    """Creates a context in which modules, other than those in
    allowed_real_modules, will not be imported but instead replaced with
    mocks.

    Manager sys.modules so that the mock modules are removed after
    the context ends.

    Allows python code to be imported and executed even when its
    dependencies are not installed.
    """

    def __init__(self, *allowed_real_modules):
        self.allowed_real_modules = set(allowed_real_modules)

    def __enter__(self):
        self.orig_sys_meta_path = sys.meta_path
        self.orig_sys_modules = sys.modules
        self.finder = DummyFinder(self.allowed_real_modules, self.orig_sys_meta_path)
        sys.modules = dict(sys.modules)
        sys.meta_path = [self.finder]

    def __exit__(self, *args):
        sys.meta_path = self.orig_sys_meta_path
        sys.modules = self.orig_sys_modules
